/**
 * This software was developed at the National Institute of Standards and
 * Technology (NIST) by employees of the Federal Government in the course
 * of their official duties. Pursuant to title 17 Section 105 of the
 * United States Code, this software is not subject to copyright protection
 * and is in the public domain. NIST assumes no responsibility  whatsoever for
 * its use by other parties, and makes no guarantees, expressed or implied,
 * about its quality, reliability, or any other characteristic.
 */

#ifndef _IREX9_H_
#define _IREX9_H_

#include <stdint.h>
#include <vector>
#include <string>


namespace IREX9
{
   const static int32_t SUCCESS = 0,
                        FAILED_FEATURE_EXTRACTION = 2,
                        ELECTIVE_REFUSAL_TO_EXTRACT_FEATURES = 2,
                        INVOLUNTARY_FAILURE_TO_PRODUCE_TEMPLATE = 4,
                        ELECTIVE_REFUSAL_TO_PRODUCE_TEMPLATE = 6,
                        CANNOT_PARSE_INPUT = 8;

   /** Eye labels. */
   typedef enum
   {
      UNDEFINED = 0,
      RIGHT     = 1,
      LEFT      = 2
   } eye_label;


   /**
    * Defines a structure that specifies a coordinate in an image.
    */
   typedef struct
   {
      /** X-coordinate (0 == leftmost) */
      uint16_t x;

      /** Y-coordinate (0 == topmost) */
      uint16_t y;
   } point;


   /**
    * Defines a structure that holds manual segmentation information for an iris sample.
    */
   typedef struct
   {
       /** Coordinate representing manual estimate of iris center. */
       point center;

       /** Vector of points outlining pupil boundary. */
       std::vector<point> pupil_boundary;

       /** Vector of points outlining limbic (iris-sclera) boundary. */
       std::vector<point> limbic_boundary;

       /** Vector of points outlining upper eyelid boundary. */
       std::vector<point> upper_eyelid_boundary;

       /** Vector of points outlining lower eyelid boundary. */
       std::vector<point> lower_eyelid_boundary;

   } iris_boundary;


   /** Defines a structure that holds a single iris with corresponding attributes. */
   typedef struct
   {
      /** 
       *  The eye label for this iris sample (0 == undefined, 1 == right eye, 2 == left eye).
       */
      eye_label eye;

      /** Image width in pixels. */
      uint16_t image_width;

      /** Image height in pixels. */
      uint16_t image_height;

      /** The wavelength with which the iris sample was acquired. */
      uint16_t wavelength;

      /** Bit depth, 8 or 24 for RGB visible spectrum images. */
      uint8_t bit_depth;

      /** Pointer to image raster data (RGBRGBRGB...\ for 24-bit images). */
      uint8_t *data;

      /** Boundary information for this iris sample, or NULL if unspecified. */
      iris_boundary* boundary;

   } iris_sample;


   /**
    * \defgroup Functions Class A (one-to-one) Functions
    */

      /** \ingroup Functions
       * Retrieves a self-assigned identifier and contact email address for the software under test.
       *
       *  \param[out] sdk_identifier A hexidecimal integer stored as a null terminated ASCII string
       *              (e.g. "01\0"). The value can be whatever the participant chooses, but must be
       *              unique for each implementation.
       *  \param[out] email_address The point of contact for the software under test, stored as a
                      null terminated ASCII string.
       *  \return
       *          Zero indicates success. Other values indicate a vendor-defined failure.
       */
      int32_t get_pid(
         std::string& sdk_identifier,
         std::string& email_address);


      /** \ingroup Functions
       * Retrieves the maximum (per-image) enrollment and search template sizes.
       *
       * These values will be used by the test harness to pre-allocate space for template data.
       * For a vector of K iris samples, the test-harness will pre-allocate K times the provided
       * value before calling convert_multiiris_to_enrollment_template() or
       * convert_multiiris_to_identification_template().
       *
       * \param[out] max_enrollment_template_size The maximum (per-image) size of an enrollment template in bytes.
       * \param[out] max_verification_template_size The maximum (per-image) size of a search template in bytes.
       *
       * \return
       *         Zero indicates success. Other values indicate a vendor-defined failure.
       */
      int32_t get_max_template_sizes(
          uint32_t& max_enrollment_template_size,
          uint32_t& max_verification_template_size);


      /** \ingroup Functions
       * Generates an enrollment template from a vector of iris samples.
       *
       * If the function returns a zero exit status, the template will be used for matching.
       * If the function returns a value of 8, NIST will debug.
       * Otherwise, a non-zero return value will indicate a failure to
       * acquire and the template will \a not be used in subsequent search operations.    
       *
       * \param[in] input_irides The iris samples from which to generate the template.
       * \param[out] template_size The size, in bytes, of the output template.
       * \param[out] enrollment_template Template generated from the iris samples. The template's format
                        is proprietary and NIST will not access any part of it other than to store it in the
                        EDB. The memory for the template will be pre-allocated by the NIST test harness. The
                        implementation shall \a not allocate this memory.
       * \return \make_template_table
       */
      int32_t convert_multiiris_to_enrollment_template(
         const std::vector<iris_sample>& input_irides,
         uint32_t& template_size,
         uint8_t* enrollment_template);


      /** \ingroup Functions
       * Generates a verification template from a vector of iris samples.
       *
       * If the function returns a zero exit status, the template will be used for matching.
       * If the function returns a value of 8, NIST will debug.
       * Otherwise, a non-zero return value will indicate a failure to
       * acquire and the template will \a not be used in subsequent search operations.
       *
       * \param[in] input_irides The iris samples from which to generate the template.
       * \param[out] template_size The size, in bytes, of the output template
       * \param[out] verification_template Template generated from the iris samples. The template's format
                        is proprietary and NIST will not access any part of it other to pass it to
                        identify_template() and possibly store it temporarily. The memory for the template will
                        be pre-allocated by the NIST test harness. The implementation shall \a not allocate
                        this memory.
       * \return \make_template_table
       * 
       * If there are multiple iris samples, then a zero status should be returned as long as feature
       * information could be extracted from at least one of the images.
       */
      int32_t convert_multiiris_to_verification_template(
         const std::vector<iris_sample>& input_irides,
         uint32_t& template_size,
         uint8_t* verification_template);


      /** \ingroup Functions
       * Searches a template against an enrollment template and produces a dissimilarity score.
       *
       * \param[in] verification_template A template generated by a call to convert_multiiris_to_verification_template().
       * \param[in] verification_template_size The size, in bytes, of the verification template.
       * \param[in] enrollment_template A template generated by a call to convert_multiiris_to_enrollment_template().
       * \param[in] enrollment_template_size The size, in bytes, of the enrollment template.
       * \param[out] dissimilarity A non-negative measure of the amount of dissimilarity between the templates.
       * \return \match_table
       */
      int32_t match_templates(
           const uint8_t* verification_template,
           const uint32_t verification_template_size,
           const uint8_t* enrollment_template,
           const uint32_t enrollment_template_size,
           double& dissimilarity);

} /* end namespace */

#endif	/*  _IREX9_H */

