/**
 * This software was developed at the National Institute of Standards and
 * Technology (NIST) by employees of the Federal Government in the course
 * of their official duties. Pursuant to title 17 Section 105 of the
 * United States Code, this software is not subject to copyright protection
 * and is in the public domain. NIST assumes no responsibility  whatsoever for
 * its use by other parties, and makes no guarantees, expressed or implied,
 * about its quality, reliability, or any other characteristic.
 */

#ifndef _IREX9_H_
#define _IREX9_H_

#include <stdint.h>
#include <vector>
#include <string>

namespace IREX9
{
   const static int32_t SUCCESS = 0,
                        INPUT_TEMPLATE_DEFECTIVE = 2,
                        FAILED_FEATURE_EXTRACTION = 2,
                        ELECTIVE_REFUSAL_TO_EXTRACT_FEATURES = 2,
                        INVOLUNTARY_FAILURE_TO_PRODUCE_TEMPLATE = 4,
                        ELECTIVE_REFUSAL_TO_PRODUCE_TEMPLATE = 6,
                        CANNOT_PARSE_INPUT = 8;

   /** Eye labels. */
   typedef enum
   {
      UNDEFINED = 0,
      RIGHT     = 1,
      LEFT      = 2
   } eye_label;


   /**
    * Defines a structure that specifies a coordinate in an image.
    */
   typedef struct
   {
      /** X-coordinate (0 == leftmost) */
      uint16_t x;

      /** Y-coordinate (0 == topmost) */
      uint16_t y;
   } point;


   /**
    * Defines a structure that holds manual segmentation information for an iris sample.
    *
    * \latexonly\highlightgreen{If boundary information is unspecified or unknown, the
    * center will be set to (-1,-1) and all vectors will be empty (i.e. have zero size).}\endlatexonly
    *
    * \latexonly\input{boundary_explanation}\endlatexonly
    *
    * \latexonly\highlightgreen{The center point is the inspecter's visual estimate of
    * the center of the iris/pupil structure. The identification of this center is likely
    * to be less robust than the manually identified boundary points.}\endlatexonly
    *
    * \latexonly\highlightgreen{The number of boundary points identified by the inspector
    * varies from one sample to the next. The inspector identified up to 12 pupil-iris
    * boundary points for a sample. When the eyelids occlude part of the pupil, they
    * sometimes prevent some pupil-iris boundary points from being identified. The points
    * are approximately equally spaced along the boundary.}\endlatexonly
    *
    * \latexonly\highlightgreen{On average the inspector identified about 12 points along
    * the outer boundary. Outer boundary points mark either limbus or eyelid boundaries.
    * Sometimes the inspector identified more than 12 points and sometimes fewer. Eyelid
    * boundary points are only identified where they overlap the iris or pupil. On average,
    * 8 limbus boundary points were identified per sample, 3 upper-eyelid boundary points
    * where identified per sample, and 1 lower-eyelid boundary point was identified per
    * sample.}\endlatexonly
    *
    * \latexonly\highlightgreen{We can make no guarantees with respect to the order of the
    * boundary points in each vector. For example, we cannot guarantee they are ordered
    * counter-clockwise around the center point. Although steps were taken during the manual
    * markup proces to minimize the chances of making a mistake, we cannot guarantee to a
    * certainty that all boundary points are properly located. NIST will fix mistakes if/when
    * it finds them.}\endlatexonly
    *
    * \latexonly\highlightgreen{The validation package will contain a manual markup of a
    * at least one iris sample for reference.}\endlatexonly
    */
   typedef struct
   {
       /** Cordinate representing manual estimate of iris center. */
       point center;

       /** Vector of points outlining pupil boundary. */
       std::vector<point> pupil_boundary;

       /** Vector of points outlining limbic (iris-sclera) boundary. */
       std::vector<point> limbic_boundary;

       /** Vector of points outlining upper eyelid boundary. */
       std::vector<point> upper_eyelid_boundary;

       /** Vector of points outlining lower eyelid boundary. */
       std::vector<point> lower_eyelid_boundary;

   } iris_boundary;


   /** Defines a structure that holds a single iris with corresponding attributes. */
   typedef struct
   {
      /** 
       *  The eye label for this iris sample (0 == undefined, 1 == right eye, 2 == left eye).
       */
      eye_label eye;

      /** Image width in pixels. */
      uint16_t image_width;

      /** Image height in pixels. */
      uint16_t image_height;

      /** The wavelength in nanometers with which the iris sample was illuminated. */
      uint16_t wavelength;

      /** Bit depth, 8 or 24 for RGB visible spectrum images. */
      uint8_t bit_depth;

      /** Pointer to image raster data (in RGBRGBRGB... format for 24-bit images). */
      uint8_t *data;

      /* Boundary information for this iris sample, or NULL if unspecified. */
      iris_boundary* boundary;

   } iris_sample;


   /** Defines a structure that holds a single candidate. */
   typedef struct
   {
      /** Indicates whether the candidate is valid (0=valid, 1-255=invalid).
        * \latexonly\highlightorange{If this value is non-zero, the values for
        * template\_id and distance\_score will be ignored.}\endlatexonly */
      uint8_t failed;

      /** Template identifier from the enrollment database. */
      uint32_t template_id;

      /** Measure of distance between the searched template and the candidate.
       *  Lower scores indicate greater similarity. The distance score must be
       *  non-negative, unless the search template is somehow broken, in which case
       * it shall be set to -1.
       */
      double distance_score;

   } candidate;


   /**
    * \defgroup Functions Class B (one-to-many) Functions
    */

   /** \ingroup Functions
    * Retrieves a self-assigned identifier and contact email address for the software under test.
    *
    *  \param[out] sdk_identifier A hexidecimal integer stored as a null terminated ASCII string
    *              (e.g. "01\0"). The value can be whatever the participant chooses, but must be
    *              unique for each implementation.
    *  \param[out] email_address The point of contact for the software under test, stored as a
                   null terminated ASCII string.
    *  \return
    *          Zero indicates success. Other values indicate a vendor-defined failure.
    */
   int32_t get_pid(
      std::string& sdk_identifier,
      std::string& email_address);

   /** \ingroup Functions
    * Retrieves the maximum (per-image) enrollment and search template sizes.
    *
    * These values will be used by the test harness to pre-allocate space for template data.
    * For a vector of K iris samples, the test-harness will pre-allocate K times the provided
    * value before calling convert_multiiris_to_enrollment_template() or
    * convert_multiiris_to_identification_template().
    *
    * \param[out] max_enrollment_template_size The maximum (per-image) size of an enrollment template in bytes.
    * \param[out] max_search_template_size The maximum (per-image) size of a search template in bytes.
    *
    * \return
    *         Zero indicates success. Other values indicate a vendor-defined failure.
    */
   int32_t get_max_template_sizes(
	   uint32_t& max_enrollment_template_size,
	   uint32_t& max_search_template_size);

   /** \ingroup Functions
    * Initialization function, called once prior to one or more calls to convert_multiiris_to_enrollment_template().
    *
    * The implementation shall tolerate execution of multiple calls to this function from different processes
    * running on the same machine. Each process may be reading and writing to the enrollment directory.
    *
    * \param[in] configuration_location Path to a \a read-only directory containing vendor-supplied configuration
    *               parameters and/or runtime data files.
    * \param[in] enrollment_directory The directory will be initially empty, but may have been initialized and
                    populated by separate invocations of the enrollment process. The software may
                    populate this folder in any manner it sees fit.
    * \param[in] num_persons The number of persons who will be enrolled in the database.
    * \param[in] num_images The number of images, summed over all identities, that will be used to build the
                    enrollment database.
    * \return \init_enroll_table
    *
    */
   int32_t initialize_enrollment_session(
      const std::string& configuration_location,
      const std::string& enrollment_directory,
      const uint32_t num_persons,
      const uint32_t num_images);


   /** \ingroup Functions
    * Generates an enrollment template from a vector of iris samples.
    *
    * If the function returns a zero exit status, the calling application will store the template in the EDB,
    * which is later be passed to finalize_enrollment(). If the function returns a value of 8, NIST will
    * debug. Otherwise, a non-zero return value will indicate a failure to enroll. The template will still be
    * added to the EDB and the manifest to ensure that an \a N person enrollment database contains \a N entries.
    * If the function crashes, NIST will include a zero-length template in the EDB and the manifest. The
    * finalization process must be able to process zero-length templates.
    *
    * \a IMPORTANT: The implementation shall not attempt to write to the enrollment directory (nor to other
    * resources) during this call. Data collected from the iris samples should be stored in the template
    * or created from the templates during the finalization step.
    *
    * \param[in] input_irides The iris samples from which to generate the template.
    * \param[out] template_size The size, in bytes, of the output template.
    * \param[out] enrollment_template Template generated from the iris samples. The template's format
                     is proprietary and NIST will not access any part of it other than to store it in the
                     EDB. The memory for the template will be pre-allocated by the NIST test harness. The
                     implementation shall \a not allocate this memory.
    * \return \make_template_table
    */
   int32_t convert_multiiris_to_enrollment_template(
      const std::vector<iris_sample>& input_irides,
      uint32_t& template_size,
      uint8_t *enrollment_template);


   /** \ingroup Functions
    * Finalization function, used to construct an enrollment database from an EDB and its manifest.
    *
    * Finalization shall be performed after all enrollment processes are complete. It should populate the
    * contents of the enrollment directory with everything that is necessary to perform searches against it.
    * This function allows post-enrollment book-keeping, normalization, and other statistical processing of
    * the generated templates. It should tolerate being called multiple times, altough subsequent calls
    * should probably not do anything.
    *
    * The format of the two input files is described in the table below.
    * The enrollment database (EDB) file stores a concatenation of the templates generated by calls to
    * convert_multiiris_to_enrollment_template() in binary format. It does not contain a header or any
    * delimiters between templates. This file can potentially be several gigabytes in size. The EDB
    * manifest is an ASCII file that stores information about each template in the EDB file. Each line
    * contains three space-delimited fields specifying the id, length, and offset of the template in the
    * EDB file. If the EDB file contains N templates, the manifest will contain N lines.
    *
    * For all intents and purposes, the template id can be regarded as a person id.
    *
    * \manifest_table
    *
    * \param[in] enrollment_directory The top-level directory in which the enrollment database will reside.
    *               The implementation will have read and write access to this directory.
    * \param[in] edb_name The path to a single \a read-only file containing the concatenated templates. The
                    implementation should extract content from this file and place it in the enrollment directory.
    * \param[in] edb_manifest_name The path to a single \a read-only file containing the EDB manifest.
    * \return \finalize_table
    */
   int32_t finalize_enrollment(
      const std::string& enrollment_directory,
      const std::string& edb_name,
      const std::string& edb_manifest_name);


   /** \ingroup Functions
    * Initialization function, to be called once prior to one or more calls to convert_multiiris_to_identification_template().
    *
    * The implementation shall tolerate execution of multiple calls to this function from different processes
    * running on the same machine.
    *
    * \param[in] configuration_location Path to a \a read-only directory containing vendor-supplied configuration
    *               parameters and/or runtime data files.
    * \param[in] enrollment_directory The top-level directory in which the enrollment data was placed when
                    finalize_enrollment() was called.
    * \param[out] expected_memsize Given the enrollment data, the implementation shall specify the expected or
                    peak memory size (in bytes) that will be used during searching.
    * \return \init_feat_extract_table
   */
   int32_t initialize_feature_extraction_session(
      const std::string& configuration_location,
      const std::string& enrollment_directory,
      uint64_t& expected_memsize);


   /** \ingroup Functions
    * Generates an identification template from a vector of iris samples.
    *
    * If the function returns a zero exit status, the template will be used for matching.
    * If the function returns a value of 8, NIST will debug.
    * Otherwise, a non-zero return value will indicate a failure to
    * acquire and the template will \a not be used in subsequent search operations.
    *
    * \param[in] input_irides The iris samples from which to generate the template.
    * \param[out] template_size The size, in bytes, of the output template
    * \param[out] identification_template Template generated from the iris samples. The template's format
                     is proprietary and NIST will not access any part of it other to pass it to
                     identify_template() and possibly store it temporarily. The memory for the template will
                     be pre-allocated by the NIST test harness. The implementation shall \a not allocate
                     this memory.
    * \return \make_template_table
    * 
    * If there are multiple iris samples, then a zero status should be returned as long as feature
    * information could be extracted from at least one of the images.
    */
   int32_t convert_multiiris_to_identification_template(
      const std::vector<iris_sample>& input_irides,
      uint32_t& template_size,
      uint8_t *identification_template);


   /** \ingroup Functions
    * Initialization function, to be called once prior to one or more calls to identify_template().
    *
    * The function may read data (e.g. templates) from the enrollment directory and load them into memory.
    *
    * \param[in] configuration_location Path to a \a read-only directory containing vendor-supplied configuration
    *               parameters and/or runtime data files.
    * \param[in] enrollment_directory The top-level directory in which the enrollment data was placed when
    *               finalize_enrollment() was called.
    * \return \ident_table
    */
   int32_t initialize_identification_session(
      const std::string& configuration_location,
      const std::string& enrollment_directory);


   /** \ingroup Functions
    * Searches a template against the enrollment database and returns a list of candidates.
    *
    * NIST will typically set the candidate list length to operationally feasible values (e.g. 20), but may
    * decide to extend it to values that approach the size of the enrollment database.
    *
    * \param[in] identification_template A template generated by a call to convert_multiiris_to_identification_template().
    * \param[in] identification_template_size The size, in bytes, of the template.
    * \param[in] candidate_list_length The length of the candidate list array.
    * \param[out] candidate_list An array (of length candidate_list_length) of pointers to candidates. 
    *                Each candidate shall be populated by the implementation and shall be sorted in ascending
    *                order of distance score (e.g. the most similar entry shall appear first).
    *                The candidate list must be populated with sensible values.
    * \return \identify_table
   */
   int32_t identify_template(
	const uint8_t *identification_template,
	const uint32_t identification_template_size,
	const uint32_t candidate_list_length,
	std::vector<candidate>& candidate_list);


} /* end namespace */

#endif	/*  _IREX9_H */

