/**
 * This software was developed at the National Institute of Standards and
 * Technology (NIST) by employees of the Federal Government in the course
 * of their official duties. Pursuant to title 17 Section 105 of the
 * United States Code, this software is not subject to copyright protection
 * and is in the public domain. NIST assumes no responsibility  whatsoever for
 * its use by other parties, and makes no guarantees, expressed or implied,
 * about its quality, reliability, or any other characteristic.
 **/
#include "irex9_classA.h"
#include "read_jpeg.h"
#include <iostream>
#include <fstream>
#include <string>
#include <limits>
#include <memory>

using namespace std;

// List of comparisons to perform. Each line lists the verification image path
// followed by the enrollment image path.
const string comparison_file = "comparisons.list";

int main(int argc, char** argv)
{
   uint32_t max_enrollment_template_size,
            max_recognition_template_size;

   // Get maximum template sizes.
   int32_t ret = IREX9::get_max_template_sizes(max_enrollment_template_size,
                                               max_recognition_template_size);
   if (ret != IREX9::SUCCESS)
   {
      cerr << "Non-zero return value from 'get_max_template_sizes': " << ret << "." << endl;
      return ret;
   }

   // Pre-allocate space for iris templates.
   unique_ptr<uint8_t[]> edata(new uint8_t[max_enrollment_template_size]),
                         vdata(new uint8_t[max_recognition_template_size]);

   // Open comparison file.
   ifstream ifs(comparison_file);
   if (!ifs)
   {
      cerr << "Failed to open stream for " << comparison_file << "." << endl;
      return 1;
   }

   // Will store file paths for verification and enrollment PGMs.
   string vfile,
          efile;

   // Will store template IDs for verification and enrollment images.
   uint32_t vid,
            eid;

   // Will store size of verification and enrollment templates after calls to 
   // convert_multiiris_to_*_template().
   uint32_t vsize,
            esize;

   // Will store iris image after call to convert_multiiris_to_*_template().
   IREX9::iris_sample is;

   // Will store dissimilarity between samples.
   double dissimilarity;

   // For each comparison.
   while (ifs >> vid >> vfile >> eid >> efile)
   {
      // Read PGM file and create iris_sample from it.
      bool success = read_jpeg(vfile, is);
      if (!success)
      {
         cerr << "Failed to load PGM file: " << vfile << "." << endl;
         return 2;
      }

      // Provide hardcoded boundary points for the given image.
      if (vfile == "images/099/S6099S00.jpg")
      {
         // Manually provide boundary coordinates for this image.
         is.boundary = new IREX9::iris_boundary;

         is.boundary->center = {284, 168};

         is.boundary->limbic_boundary = { {408, 169}, {392, 231}, {349, 274}, {288, 290},
                                          {225, 277}, {177, 233}, {159, 174} };

         is.boundary->pupil_boundary = { {283, 121}, {306, 128}, {322, 142}, {329, 167},
                                         {323, 189}, {306, 206}, {284, 211}, {262, 206},
                                         {244, 187}, {239, 167}, {246, 146}, {259, 129} };

         is.boundary->upper_eyelid_boundary = { {191, 143}, {234, 119}, {283, 107}, {332, 105},
                                                {389, 112} };

         is.boundary->lower_eyelid_boundary = { }; // no lower-eyelid occlusion
      }

      vector<IREX9::iris_sample> multi_iris(1, is);
      vsize = 0;

      // Create verification template from iris sample.
      ret = IREX9::convert_multiiris_to_verification_template(multi_iris, vsize, vdata.get());
      if (ret == IREX9::CANNOT_PARSE_INPUT)
      {
         cerr << "Failed to generate template for " << vfile << ", " << ret << "." << endl;
         return ret;
      }
      else if (ret != IREX9::SUCCESS)
      {
         // Should still produce a matchable template.
      }

      // Free pixel data in iris sample;
      delete[] is.data;
      is.data = NULL;

      if (is.boundary != NULL)
      {
         delete is.boundary;
         is.boundary = NULL;
      }


      // Read PGM file and create iris_sample from it.
      success = read_jpeg(efile, is);
      if (!success)
      {
         cerr << "Failed to load PGM file: " << efile << "." << endl;
         return 3;
      }

      multi_iris[0] = is;
      esize = 0;

      // Create enrollment template from iris sample.
      ret = IREX9::convert_multiiris_to_enrollment_template(multi_iris, esize, edata.get());
      if (ret == IREX9::CANNOT_PARSE_INPUT)
      {
         cerr << "Failed to generate template for " << efile << ", " << ret << "." << endl;
         return ret;
      }
      else if (ret != IREX9::SUCCESS)
      {
         // Should still produce a matchable template.
      }

      // Free pixel data in iris sample;
      delete[] is.data;
      is.data = NULL;

      if (is.boundary != NULL)
      {
         delete is.boundary;
         is.boundary = NULL;
      }

      // Compare templates.
      ret = IREX9::match_templates(vdata.get(), vsize, edata.get(), esize, dissimilarity);
      if (ret == IREX9::FAILED_FEATURE_EXTRACTION)
      {
         dissimilarity = numeric_limits<double>::infinity();
      }
      else if (ret != IREX9::SUCCESS)
      {
         cerr << "Error during call to match_templates(): " << ret << endl;
         return ret;
      }

      // Write to standard output.
      cout << vid << " " << eid << " " << ret << " " << dissimilarity << endl;
   }

   if (ifs.bad())
   {
      cerr << "Error while reading " << comparison_file << "." << endl;
      return 4;
   }

   return 0;
}

