/**
 * This software was developed at the National Institute of Standards and
 * Technology (NIST) by employees of the Federal Government in the course
 * of their official duties. Pursuant to title 17 Section 105 of the
 * United States Code, this software is not subject to copyright protection
 * and is in the public domain. NIST assumes no responsibility  whatsoever for
 * its use by other parties, and makes no guarantees, expressed or implied,
 * about its quality, reliability, or any other characteristic.
 **/
#include "irex9_classB.h"
#include "read_jpeg.h"

#include <iostream>
#include <fstream>
#include <string>
#include <memory>

using namespace std;

const string configuration_dir = ".",
             enrollment_file   = "enrollment.list",
             verification_file = "verification.list",
             enroll_dir        = "enroll_dir",
             edb_name          = enroll_dir + "/edb",
             edb_manifest_name = enroll_dir + "/edb.manifest";

const uint32_t cand_length = 32;


int main(int argc, char** argv)
{
   uint32_t max_enrollment_template_size,
            max_recognition_template_size;

   // Read in maximum template sizes.
   int32_t ret = IREX9::get_max_template_sizes(max_enrollment_template_size,
                                               max_recognition_template_size);
   if (ret != IREX9::SUCCESS)
   {
      cerr << "Non-zero return value from 'get_max_template_sizes': " << ret << "." << endl;
      return ret;
   }


   ret = IREX9::initialize_enrollment_session(configuration_dir, enroll_dir, cand_length, cand_length);
   if (ret != IREX9::SUCCESS)
   {
      cerr << "Non-zero return value from 'initialize_enrollment_session': " << ret << "." << endl;
      return ret;
   }


   // Open enrollment file.
   ifstream efs(enrollment_file);
   if (!efs)
   {
      cerr << "Failed to open stream for " << enrollment_file << "." << endl;
      return 1;
   }

   // Open edb file.
   ofstream edb_stream(edb_name, ios::binary);
   if (!edb_stream)
   {
      cerr << "Failed to open stream for " << edb_name << "." << endl;
      return 1;
   }

   // Open edb manifest file.
   ofstream edb_manifest_stream(edb_manifest_name);
   if (!edb_manifest_stream)
   {
      cerr << "Failed to open stream for " << edb_manifest_name << "." << endl;
      return 1;
   }


   // Pre-allocate space for iris templates.
   vector<uint8_t> template_data(max_enrollment_template_size);

   uint32_t template_size;
   uint32_t template_id;
   string image_path;

   // Foreach enrollment image.
   while (efs >> template_id >> image_path)
   {
      IREX9::iris_sample is;

      // Read PGM file and create iris_sample from it.
      bool success = read_jpeg(image_path, is);
      if (!success)
      {
         cerr << "Failed to load PGM file: " << image_path << "." << endl;
         return 1;
      }

      vector<IREX9::iris_sample> multi_iris(1, is);
      template_size = 0;

      // Create enrollment template from iris sample.
      ret = IREX9::convert_multiiris_to_enrollment_template(multi_iris, template_size, template_data.data());
      if (ret == IREX9::CANNOT_PARSE_INPUT)
      {
         cerr << "Failed to generate template for " << image_path << ", " << ret << "." << endl;
         return ret;
      }
      else if (ret != IREX9::SUCCESS)
      {
         // Should still produce a matchable template.
      }

      // Free pixel data in iris sample;
      delete[] is.data;
      is.data = NULL;

      if (is.boundary != NULL)
      {
         delete is.boundary;
         is.boundary = NULL;
      }


      // Save to manifest and EDB files.
      edb_manifest_stream << template_id << " " << template_size << " " << edb_stream.tellp() << endl;
      edb_stream.write((char*)template_data.data(), template_size);

      if (edb_stream.fail() || edb_stream.bad())
      {
         cerr << "Error writing to EDB stream." << endl;
         return 2;
      }
   }

   edb_stream.close();
   edb_manifest_stream.close();

   if (efs.bad())
   {
      cerr << "Error while reading " << enrollment_file << endl;
      return 2;
   }

   // Finalize Enrollment.
   ret = IREX9::finalize_enrollment(enroll_dir, edb_name, edb_manifest_name);
   if (ret != IREX9::SUCCESS)
   {
      cerr << "Non-zero return value from 'finalize_enrollment': " << ret << "." << endl;
      return ret;
   }

   template_data.resize(max_recognition_template_size);

   uint64_t expected_memsize;

   // Initialize Feature Extraction.
   ret = IREX9::initialize_feature_extraction_session(configuration_dir, enroll_dir, expected_memsize);
   if (ret != IREX9::SUCCESS)
   {
      cerr << "Non-zero return value from 'initialize_feature_extraction_session': " << ret << "." << endl;
      return ret;
   }

   // Initialize identification session.
   ret = IREX9::initialize_identification_session(configuration_dir, enroll_dir);
   if (ret != IREX9::SUCCESS)
   {
      cerr << "Non-zero return value from 'initialize_identification_session': " << ret << "." << endl;
      return ret;
   }

   // Open verification file.
   ifstream vfs(verification_file);
   if (!vfs)
   {
      cerr << "Failed to open stream for " << verification_file << "." << endl;
      return 1;
   }


   // Foreach probe file.
   while (vfs >> template_id >> image_path)
   {
      IREX9::iris_sample is;

      // Read PGM file and create iris_sample from it.
      bool success = read_jpeg(image_path, is);
      if (!success)
      {
         cerr << "Failed to load PGM file: " << image_path << "." << endl;
         return 2;
      }

      vector<IREX9::iris_sample> multi_iris(1, is);
      template_size = 0;

      // Create verification template from iris sample.
      ret = IREX9::convert_multiiris_to_identification_template(multi_iris, template_size, template_data.data());
      if (ret == IREX9::CANNOT_PARSE_INPUT)
      {
         cerr << "Failed to generate template for " << image_path << ", " << ret << "." << endl;
         return ret;
      }

      // Free pixel data in iris sample;
      delete[] is.data;
      is.data = NULL;

      if (is.boundary != NULL)
      {
         delete is.boundary;
         is.boundary = NULL;
      }

      // If convert_multiiris_to_identification_template() was successful.
      if (ret == IREX9::SUCCESS)
      {
         vector<IREX9::candidate> candidates;
         candidates.reserve(cand_length);

         // Compare templates.
         ret = IREX9::identify_template(template_data.data(), template_size, cand_length, candidates);
         if (ret != IREX9::SUCCESS)
         {
            cerr << "Error during call to identify_template(): " << ret << endl;
            return ret;
         }

         // Write to standard output.
         for (const auto& i : candidates)
            cout << template_id << " " << i.template_id << " "
                 << static_cast<int>(i.failed) << " " << i.distance_score << endl;
      }
   }

   return 0;
}

